/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2004 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.0 of the PHP license,       |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_0.txt.                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Author: Sara Golemon <pollita@php.net>                               |
  +----------------------------------------------------------------------+
*/

/* $Id: classkit.c,v 1.9 2004/08/26 00:57:58 pollita Exp $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "ext/standard/info.h"
#include "ext/standard/php_string.h"
#include "php_classkit.h"

/* Generic Constants */
#define CLASSKIT_VERSION	"0.3"

/* Aggregation Flags */
#define CLASSKIT_AGGREGATE_OVERRIDE	0x0001

/* {{{ classkit_functions[]
 *
 * Every user visible function must have an entry in classkit_functions[].
 */
function_entry classkit_functions[] = {
	PHP_FE(classkit_method_add,					NULL)
	PHP_FE(classkit_method_redefine,			NULL)
	PHP_FE(classkit_method_remove,				NULL)
	PHP_FE(classkit_method_rename,				NULL)
	PHP_FE(classkit_method_copy,				NULL)
	PHP_FE(classkit_aggregate_methods,			NULL)
	PHP_FE(classkit_import,						NULL)
	{NULL, NULL, NULL}
};
/* }}} */

/* {{{ classkit_module_entry
 */
zend_module_entry classkit_module_entry = {
#if ZEND_MODULE_API_NO >= 20010901
	STANDARD_MODULE_HEADER,
#endif
	"classkit",
	classkit_functions,
	PHP_MINIT(classkit),
	NULL, /* MSHUTDOWN */
	NULL, /* RINIT */
	NULL, /* RSHUTDOWN */
	PHP_MINFO(classkit),
#if ZEND_MODULE_API_NO >= 20010901
	CLASSKIT_VERSION, 
#endif
	STANDARD_MODULE_PROPERTIES
};
/* }}} */

#ifdef COMPILE_DL_CLASSKIT
ZEND_GET_MODULE(classkit)
#endif

/* {{{ PHP_MINFO_FUNCTION
 */
PHP_MINFO_FUNCTION(classkit)
{
	php_info_print_table_start();
	php_info_print_table_header(2, "classkit support", "enabled");
	php_info_print_table_header(2, "version", CLASSKIT_VERSION);
	php_info_print_table_end();

}
/* }}} */

#ifdef ZEND_ENGINE_2
#define php_classkit_locate_scope(ce, fe, methodname, methodname_len)	fe->common.scope
#else

/* {{{ _php_classkit_locate_scope
	ZendEngine 1 hack to determine a function's scope */
static zend_class_entry *_php_classkit_locate_scope(zend_class_entry *ce, zend_function *fe, char *methodname, int methodname_len)
{
	zend_class_entry *current = ce->parent, *top = ce;
	zend_function *func;

	while (current) {
		if (zend_hash_find(&current->function_table, methodname, methodname_len + 1, (void **)&func) == FAILURE) {
			/* Not defined at this point (or higher) */
			return top;
		}
		if (func->op_array.opcodes != fe->op_array.opcodes) {
			/* Different function above this point */
			return top;
		}
		/* Same opcodes */
		top = current;

		current = current->parent;
	}

	return top;
}
/* }}} */

#define php_classkit_locate_scope(ce, fe, methodname, methodname_len)	_php_classkit_locate_scope((ce), (fe), (methodname), (methodname_len))
#endif

/* {{{ php_classkit_check_call_stack
 */
static int php_classkit_check_call_stack(zend_op_array *op_array TSRMLS_DC)
{
	zend_execute_data *ptr;

	ptr = EG(current_execute_data);

	while (ptr) {
		if (ptr->op_array->opcodes == op_array->opcodes) {
			return FAILURE;
		}
		ptr = ptr->prev_execute_data;
	}

	return SUCCESS;
}
/* }}} */

/* {{{ php_classkit_fetch_class
 */
static int php_classkit_fetch_class(char *classname, int classname_len, zend_class_entry **pce TSRMLS_DC)
{
	zend_class_entry *ce;

#ifdef ZEND_ENGINE_2
	zend_class_entry **ze;

	if (zend_hash_find(EG(class_table), classname, classname_len + 1, (void**)&ze) == FAILURE ||
		!ze || !*ze) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "class %s not found", classname);
		return FAILURE;
	}
	ce = *ze;
#else
	if (zend_hash_find(EG(class_table), classname, classname_len + 1, (void**)&ce) == FAILURE ||
		!ce) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "class %s not found", classname);
		return FAILURE;
	}
#endif

	if (ce->type != ZEND_USER_CLASS) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "class %s is not a user-defined class", classname);
		return FAILURE;
	}

	*pce = ce;

	return SUCCESS;
}
/* }}} */

/* {{{ php_classkit_fetch_class_method
 */
static int php_classkit_fetch_class_method(char *classname, int classname_len, char *methodname, int methodname_len,
											zend_class_entry **pce, zend_function **pfe TSRMLS_DC)
{
	zend_class_entry *ce;
	zend_function *fe;

	if (php_classkit_fetch_class(classname, classname_len, &ce TSRMLS_CC) == FAILURE) {
		return FAILURE;
	}

	if (zend_hash_find(&ce->function_table, methodname, methodname_len + 1, (void**)&fe) == FAILURE ||
		!fe) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Method %s::%s not found", classname, methodname);
		return FAILURE;
	}

	if (fe->type != ZEND_USER_FUNCTION) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "%s::%s is not a user defined method", classname, methodname);
		return FAILURE;
	}

	*pce = ce;
	*pfe = fe;

	return SUCCESS;
}
/* }}} */

#define CLASSKIT_TEMP_FUNCNAME	"__classkit_temporary_function__"
/* {{{ php_classkit_generate_lambda_method
	Heavily borrowed from ZEND_FUNCTION(create_function) */
static int php_classkit_generate_lambda_method(char *arguments, int arguments_len, char *phpcode, int phpcode_len, zend_function **pfe TSRMLS_DC) 
{
	char *eval_code, *eval_name;
	int eval_code_length;

	eval_code_length = sizeof("function " CLASSKIT_TEMP_FUNCNAME) + arguments_len + 4 + phpcode_len;
	eval_code = (char*)emalloc(eval_code_length);
	snprintf(eval_code, eval_code_length, "function " CLASSKIT_TEMP_FUNCNAME "(%s){%s}", arguments, phpcode);
	eval_name = zend_make_compiled_string_description("classkit runtime-created function" TSRMLS_CC);
	if (zend_eval_string(eval_code, NULL, eval_name TSRMLS_CC) == FAILURE) {
		efree(eval_code);
		efree(eval_name);
		return FAILURE;
	}
	efree(eval_code);
	efree(eval_name);

	if (zend_hash_find(EG(function_table), CLASSKIT_TEMP_FUNCNAME, sizeof(CLASSKIT_TEMP_FUNCNAME), (void **)pfe) == FAILURE) {
		php_error_docref(NULL TSRMLS_CC, E_ERROR, "Unexpected inconsistency during create_function");
		return FAILURE;
	}

	return SUCCESS;
}
/* }}} */

#ifdef ZEND_ENGINE_2
/* Why doesn't ZE2 define this? */
#define zend_hash_add_or_update(ht, arKey, nKeyLength, pData, pDataSize, pDest, flag) \
		_zend_hash_add_or_update((ht), (arKey), (nKeyLength), (pData), (pDataSize), (pDest), (flag) ZEND_FILE_LINE_CC)
#endif

/* {{{ php_classkit_clean_children
	Scan the class_table for children of the class just updated */
static int php_classkit_clean_children(zend_class_entry **ze, int num_args, va_list args, zend_hash_key *hash_key)
{
	zend_class_entry *ancestor_class =  va_arg(args, zend_class_entry*);
	zend_class_entry *parent_class =  va_arg(args, zend_class_entry*);
	zend_class_entry *ce, *scope;
	char *fname = va_arg(args, char*);
	int fname_len = va_arg(args, int);
	zend_function *cfe = NULL;
	TSRMLS_FETCH();

#ifdef ZEND_ENGINE_2
	ce = *ze;
#else
	ce = (zend_class_entry *)ze;
#endif

	if (ce->parent != parent_class) {
		/* Not a child, ignore */
		return ZEND_HASH_APPLY_KEEP;
	}

	if (zend_hash_find(&ce->function_table, fname, fname_len + 1, (void**)&cfe) == SUCCESS) {
		scope = php_classkit_locate_scope(ce, cfe, fname, fname_len);
		if (scope != ancestor_class) {
			/* This method was defined below our current level, leave it be */
			return ZEND_HASH_APPLY_KEEP;
		}
	}

	if (!cfe) {
		/* Odd.... nothing to destroy.... */
		return ZEND_HASH_APPLY_KEEP;
	}

	/* Process children of this child */
	zend_hash_apply_with_arguments(EG(class_table), (apply_func_args_t)php_classkit_clean_children, 4, ancestor_class, ce, fname, fname_len);

	zend_hash_del(&ce->function_table, fname, fname_len + 1);

	return ZEND_HASH_APPLY_KEEP;
}
/* }}} */

/* {{{ php_classkit_update_children
	Scan the class_table for children of the class just updated */
static int php_classkit_update_children(zend_class_entry **ze, int num_args, va_list args, zend_hash_key *hash_key)
{
	zend_class_entry *ancestor_class =  va_arg(args, zend_class_entry*);
	zend_class_entry *parent_class =  va_arg(args, zend_class_entry*);
	zend_class_entry *ce, *scope;
	zend_function *fe =  va_arg(args, zend_function*);
	char *fname = va_arg(args, char*);
	int fname_len = va_arg(args, int);
	zend_function *cfe = NULL;
	TSRMLS_FETCH();

#ifdef ZEND_ENGINE_2
	ce = *ze;
#else
	ce = (zend_class_entry *)ze;
#endif

	if (ce->parent != parent_class) {
		/* Not a child, ignore */
		return ZEND_HASH_APPLY_KEEP;
	}

	if (zend_hash_find(&ce->function_table, fname, fname_len + 1, (void**)&cfe) == SUCCESS) {
		scope = php_classkit_locate_scope(ce, cfe, fname, fname_len);
		if (scope != ancestor_class) {
			/* This method was defined below our current level, leave it be */
			return ZEND_HASH_APPLY_KEEP;
		}
	}

	/* Process children of this child */
	zend_hash_apply_with_arguments(EG(class_table), (apply_func_args_t)php_classkit_update_children, 5, ancestor_class, ce, fe, fname, fname_len);

	function_add_ref(fe);
	if (zend_hash_add_or_update(&ce->function_table, fname, fname_len + 1, fe, sizeof(zend_function), NULL, cfe ? HASH_UPDATE : HASH_ADD) ==  FAILURE) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Error updating child class");
		return ZEND_HASH_APPLY_KEEP;
	}

	return ZEND_HASH_APPLY_KEEP;
}
/* }}} */

/* {{{ php_classkit_method_add_or_update
 */
static void php_classkit_method_add_or_update(INTERNAL_FUNCTION_PARAMETERS, int add_or_update)
{
	char *classname, *methodname, *arguments, *phpcode;
	int classname_len, methodname_len, arguments_len, phpcode_len;
	zend_class_entry *ce, *ancestor_class = NULL;
	zend_function func, *fe;
#ifdef ZEND_ENGINE_2
	long flags = ZEND_ACC_PUBLIC;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ssss|l",  &classname, &classname_len,
																	&methodname, &methodname_len,
																	&arguments, &arguments_len,
																	&phpcode, &phpcode_len,
																	&flags) == FAILURE) {
#else
	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ssss", &classname, &classname_len,
																 &methodname, &methodname_len,
																 &arguments, &arguments_len,
																 &phpcode, &phpcode_len) == FAILURE) {
#endif
		RETURN_FALSE;
	}

	if (!classname_len || !methodname_len) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Empty parameter given");
		RETURN_FALSE;
	}

	php_strtolower(classname, classname_len);
	php_strtolower(methodname, methodname_len);

	if (add_or_update == HASH_UPDATE) {
		if (php_classkit_fetch_class_method(classname, classname_len, methodname, methodname_len, &ce, &fe TSRMLS_CC) == FAILURE) {
			RETURN_FALSE;
		}
		ancestor_class = php_classkit_locate_scope(ce, fe, methodname, methodname_len);

		if (php_classkit_check_call_stack(&fe->op_array TSRMLS_CC) == FAILURE) {
			php_error_docref(NULL TSRMLS_CC, E_WARNING, "Cannot redefine a method while that method is active.");
			RETURN_FALSE;
		}
	} else {
		if (php_classkit_fetch_class(classname, classname_len, &ce TSRMLS_CC) == FAILURE) {
			RETURN_FALSE;
		}
		ancestor_class = ce;
	}

	if (php_classkit_generate_lambda_method(arguments, arguments_len, phpcode, phpcode_len, &fe TSRMLS_CC) == FAILURE) {
		RETURN_FALSE;
	}

	func = *fe;
	function_add_ref(&func);
	efree(func.common.function_name);
	func.common.function_name = estrndup(methodname, methodname_len);
#ifdef ZEND_ENGINE_2
	func.common.scope = ce;

	if (flags & ZEND_ACC_PRIVATE) {
		func.common.fn_flags &= ~ZEND_ACC_PPP_MASK;
		func.common.fn_flags |= ZEND_ACC_PRIVATE;
	} else if (flags & ZEND_ACC_PROTECTED) {
		func.common.fn_flags &= ~ZEND_ACC_PPP_MASK;
		func.common.fn_flags |= ZEND_ACC_PROTECTED;
	}

	func.common.fn_flags |= ZEND_ACC_ALLOW_STATIC;
#endif

	zend_hash_apply_with_arguments(EG(class_table), (apply_func_args_t)php_classkit_update_children, 5, ancestor_class, ce, &func, methodname, methodname_len);

	if (zend_hash_add_or_update(&ce->function_table, methodname, methodname_len + 1, &func, sizeof(zend_function), NULL, add_or_update) == FAILURE) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Unable to add method to class");
		RETURN_FALSE;
	}

	if (zend_hash_del(EG(function_table), CLASSKIT_TEMP_FUNCNAME, sizeof(CLASSKIT_TEMP_FUNCNAME)) == FAILURE) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Unable to remove temporary function entry");
		RETURN_FALSE;
	}

	if (zend_hash_find(&ce->function_table, methodname, methodname_len + 1, (void**)&fe) == FAILURE ||
		!fe) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Unable to locate newly added method");
		RETURN_FALSE;
	}

#ifdef ZEND_ENGINE_2
	if (strcmp(methodname, ce->name) == 0 ||
		strcmp(methodname, "__construct") == 0)		ce->constructor = fe;
	else if (strcmp(methodname, "__destruct") == 0)	ce->destructor	= fe;
	else if (strcmp(methodname, "__clone") == 0)	ce->clone		= fe;
	else if (strcmp(methodname, "__get") == 0)		ce->__get		= fe;
	else if (strcmp(methodname, "__set") == 0)		ce->__set		= fe;
	else if (strcmp(methodname, "__call") == 0)		ce->__call		= fe;

	if (strcmp(methodname, ce->name) == 0 ||
		strcmp(methodname, "__construct") == 0)		{
		ce->constructor = fe;
		fe->common.fn_flags |= ZEND_ACC_CTOR;
	} else if (strcmp(methodname, "__destruct") == 0)	{
		ce->destructor = fe;
		fe->common.fn_flags |= ZEND_ACC_DTOR;
	} else if (strcmp(methodname, "__clone") == 0)	{
		ce->clone = fe;
		fe->common.fn_flags |= ZEND_ACC_CLONE;
	} else if (strcmp(methodname, "__get") == 0)	ce->__get		= fe;
	  else if (strcmp(methodname, "__set") == 0)	ce->__set		= fe;
	  else if (strcmp(methodname, "__call") == 0)	ce->__call		= fe;
#endif 

	RETURN_TRUE;
}
/* }}} */

/* {{{ php_classkit_function_copy_ctor
 Duplicate structures in an op_array where necessary to make an outright duplicate */
static void php_classkit_function_copy_ctor(zend_function *fe, char *newname)
{
	zend_op *opcode_copy;
	int i;

	if (fe->op_array.static_variables) {
		HashTable *tmpHash;
		zval tmpZval;

		ALLOC_HASHTABLE(tmpHash);
		zend_hash_init(tmpHash, 2, NULL, ZVAL_PTR_DTOR, 0);
		zend_hash_copy(tmpHash, fe->op_array.static_variables, ZVAL_COPY_CTOR, &tmpZval, sizeof(zval));
		fe->op_array.static_variables = tmpHash;
	}

	fe->op_array.refcount = emalloc(sizeof(zend_uint));
	*(fe->op_array.refcount) = 1;

	opcode_copy = safe_emalloc(sizeof(zend_op), fe->op_array.last, 0);
	for(i = 0; i < fe->op_array.last; i++) {
		opcode_copy[i] = fe->op_array.opcodes[i];

		if (opcode_copy[i].op1.op_type == IS_CONST) {
			zval_copy_ctor(&opcode_copy[i].op1.u.constant);
#ifdef ZEND_ENGINE_2
		} else
		if (opcode_copy[i].op1.u.jmp_addr >= fe->op_array.opcodes &&
			opcode_copy[i].op1.u.jmp_addr <  fe->op_array.opcodes + fe->op_array.last) {
			opcode_copy[i].op1.u.jmp_addr =  opcode_copy + (fe->op_array.opcodes[i].op1.u.jmp_addr - fe->op_array.opcodes);
#endif
		}

		if (opcode_copy[i].op2.op_type == IS_CONST) {
			zval_copy_ctor(&opcode_copy[i].op2.u.constant);
#ifdef ZEND_ENGINE_2
		} else
		if (opcode_copy[i].op2.u.jmp_addr >= fe->op_array.opcodes &&
			opcode_copy[i].op2.u.jmp_addr <  fe->op_array.opcodes + fe->op_array.last) {
			opcode_copy[i].op2.u.jmp_addr =  opcode_copy + (fe->op_array.opcodes[i].op2.u.jmp_addr - fe->op_array.opcodes);
#endif
		}
	}
	fe->op_array.opcodes = opcode_copy;

	if (newname) {
		fe->op_array.function_name = newname;
	} else {
		fe->op_array.function_name = estrdup(fe->op_array.function_name);
	}

#ifdef ZEND_ENGINE_2
	fe->op_array.prototype = fe;

	if (fe->op_array.arg_info) {
		zend_arg_info *tmpArginfo;

		tmpArginfo = safe_emalloc(sizeof(zend_arg_info), fe->op_array.num_args, 0);
		for(i = 0; i < fe->op_array.num_args; i++) {
			tmpArginfo[i] = fe->op_array.arg_info[i];
			tmpArginfo[i].name = estrndup(tmpArginfo[i].name, tmpArginfo[i].name_len);
			if (tmpArginfo[i].class_name) {
				tmpArginfo[i].class_name = estrndup(tmpArginfo[i].class_name, tmpArginfo[i].class_name_len);
			}
		}
		fe->op_array.arg_info = tmpArginfo;
	}

	fe->op_array.doc_comment = estrndup(fe->op_array.doc_comment, fe->op_array.doc_comment_len);
	fe->op_array.try_catch_array = (zend_try_catch_element*)estrndup((char*)fe->op_array.try_catch_array, sizeof(zend_try_catch_element) * fe->op_array.last_try_catch);
#endif
	fe->op_array.brk_cont_array = (zend_brk_cont_element*)estrndup((char*)fe->op_array.brk_cont_array, sizeof(zend_brk_cont_element) * fe->op_array.last_brk_cont);

	fe->op_array.start_op = fe->op_array.opcodes;
}
/* }}}} */

/* {{{ php_classkit_method_copy
 */
static int php_classkit_method_copy(char *dclass, int dclass_len, char *dfunc, int dfunc_len,
									char *sclass, int sclass_len, char *sfunc, int sfunc_len TSRMLS_DC)
{
	zend_class_entry *dce, *sce;
	zend_function dfe, *sfe;

	if (php_classkit_fetch_class_method(sclass, sclass_len, sfunc, sfunc_len, &sce, &sfe TSRMLS_CC) == FAILURE) {
		return FAILURE;
	}

	if (php_classkit_fetch_class(dclass, dclass_len, &dce TSRMLS_CC) == FAILURE) {
		return FAILURE;
	}

	if (zend_hash_exists(&dce->function_table, dfunc, dfunc_len + 1)) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Destination method %s::%s() already exists", dclass, dfunc);
		return FAILURE;
	}

	dfe = *sfe;
	php_classkit_function_copy_ctor(&dfe, estrndup(dfunc, dfunc_len));
#ifdef ZEND_ENGINE_2
	dfe.common.scope = dce;
#endif

	if (zend_hash_add(&dce->function_table, dfunc, dfunc_len + 1, &dfe, sizeof(zend_function), NULL) == FAILURE) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Error adding method to class %s::%s()", dclass, dfunc);
		return FAILURE;
	}

	zend_hash_apply_with_arguments(EG(class_table), (apply_func_args_t)php_classkit_update_children, 5, dce, dce, &dfe, dfunc, dfunc_len);

	return SUCCESS;
}
/* }}} */

/* {{{ proto bool classkit_method_add(string classname, string methodname, string args, string code[, long flags])
	Add a method to an already defined class */
PHP_FUNCTION(classkit_method_add)
{
	php_classkit_method_add_or_update(INTERNAL_FUNCTION_PARAM_PASSTHRU, HASH_ADD);
}
/* }}} */

/* {{{ proto bool classkit_method_redefine(string classname, string methodname, string args, string code[, long flags])
	Redefine an already defined class method */
PHP_FUNCTION(classkit_method_redefine)
{
	php_classkit_method_add_or_update(INTERNAL_FUNCTION_PARAM_PASSTHRU, HASH_UPDATE);
}
/* }}} */

/* {{{ proto bool classkit_method_remove(string classname, string methodname)
	Remove a class method */
PHP_FUNCTION(classkit_method_remove)
{
	char *classname, *methodname;
	int classname_len, methodname_len;
	zend_class_entry *ce;
	zend_function *fe;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ss", &classname, &classname_len,
															   &methodname, &methodname_len) == FAILURE) {
		RETURN_FALSE;
	}

	if (!classname_len || !methodname_len) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Empty parameter given");
		RETURN_FALSE;
	}

	php_strtolower(classname, classname_len);
	php_strtolower(methodname, methodname_len);

	if (php_classkit_fetch_class_method(classname, classname_len, methodname, methodname_len, &ce, &fe TSRMLS_CC) == FAILURE) {
		RETURN_FALSE;
	}

	if (php_classkit_check_call_stack(&fe->op_array TSRMLS_CC) == FAILURE) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Cannot remove a method while it is active");
		RETURN_FALSE;
	}

	zend_hash_apply_with_arguments(EG(class_table), (apply_func_args_t)php_classkit_clean_children, 4, php_classkit_locate_scope(ce, fe, methodname, methodname_len), ce, methodname, methodname_len);

	if (zend_hash_del(&ce->function_table, methodname, methodname_len + 1) == FAILURE) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Error dereferencing old method");
		RETURN_FALSE;
	}

	RETURN_TRUE;
}
/* }}} */

/* {{{ proto bool classkit_method_rename(string classname, string methodname, string newname)
	Rename a class method */
PHP_FUNCTION(classkit_method_rename)
{
	char *classname, *methodname, *newname;
	int classname_len, methodname_len, newname_len;
	zend_class_entry *ce;
	zend_function *fe, tmp;
	zend_class_entry *scope;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sss", &classname, &classname_len,
																&methodname, &methodname_len,
																&newname, &newname_len) == FAILURE) {
		RETURN_FALSE;
	}

	if (!classname_len || !methodname_len || !newname_len) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Empty parameter given");
		RETURN_FALSE;
	}

	php_strtolower(classname, classname_len);
	php_strtolower(methodname, methodname_len);
	php_strtolower(newname, newname_len);

	if (strcmp(methodname, newname) == 0) {
		/* Nothing to do */
		RETURN_FALSE;
	}

	if (php_classkit_fetch_class_method(classname, classname_len, methodname, methodname_len, &ce, &fe TSRMLS_CC) == FAILURE) {
		RETURN_FALSE;
	}

	scope = php_classkit_locate_scope(ce, fe, methodname, methodname_len);
	if (ce != scope) {
		/* Inherited method: Only rename original methods please.  You're welcome to copy instead though */
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Method can only be renamed in its originating class(class %s).", 
													(scope && scope->name) ? scope->name : "Unknown");
		RETURN_FALSE;
	}

	if (php_classkit_check_call_stack(&fe->op_array TSRMLS_CC) == FAILURE) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Cannot rename a method while it is active");
		RETURN_FALSE;
	}

	if (zend_hash_exists(&ce->function_table, newname, newname_len + 1)) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "method %s::%s already exists", classname, newname);
		RETURN_FALSE;
	}

	tmp = *fe;
	/* Reuse this function in a new identity */
	efree(tmp.common.function_name);
	function_add_ref(&tmp);
	tmp.common.function_name = estrndup(newname, newname_len);

#ifdef ZEND_ENGINE_2
	tmp.common.scope = ce;
#endif

	zend_hash_apply_with_arguments(EG(class_table), (apply_func_args_t)php_classkit_update_children, 5, scope, ce, &tmp, newname, newname_len);
	zend_hash_apply_with_arguments(EG(class_table), (apply_func_args_t)php_classkit_clean_children, 4, scope, ce, methodname, methodname_len);

	if (zend_hash_add(&ce->function_table, newname, newname_len + 1, &tmp, sizeof(zend_function), NULL) == FAILURE) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Error adding new method name");
		RETURN_FALSE;
	}

	/* This delete removes the reference to the function_name freed above
		When the final reference to these opcodes dies, it'll free the newly alloc'd function_name instead */
	if (zend_hash_del(&ce->function_table, methodname, methodname_len + 1) == FAILURE) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Error dereferencing old method");
		RETURN_FALSE;
	}

#ifdef ZEND_ENGINE_2
	if (fe == ce->constructor) {
		ce->constructor = NULL;
		fe->common.fn_flags &= ~ZEND_ACC_CTOR;
	}
	if (fe == ce->destructor) {
		ce->destructor = NULL;
		fe->common.fn_flags &= ~ZEND_ACC_DTOR;
	}
	if (fe == ce->clone) {
		ce->clone = NULL;
		fe->common.fn_flags &= ~ZEND_ACC_CLONE;
	}
	if (fe == ce->__get)		ce->__get		= NULL;
	if (fe == ce->__set)		ce->__set		= NULL;
	if (fe == ce->__call)		ce->__call		= NULL;

	if (strcmp(newname, ce->name) == 0 ||
		strcmp(newname, "__construct") == 0)		{
		ce->constructor = fe;
		fe->common.fn_flags |= ZEND_ACC_CTOR;
	} else if (strcmp(newname, "__destruct") == 0)	{
		ce->destructor = fe;
		fe->common.fn_flags |= ZEND_ACC_DTOR;
	} else if (strcmp(newname, "__clone") == 0)	{
		ce->clone = fe;
		fe->common.fn_flags |= ZEND_ACC_CLONE;
	} else if (strcmp(newname, "__get") == 0)	ce->__get		= fe;
	  else if (strcmp(newname, "__set") == 0)	ce->__set		= fe;
	  else if (strcmp(newname, "__call") == 0)	ce->__call		= fe;
#endif

	RETURN_TRUE;
}
/* }}} */

/* {{{ proto bool classkit_method_copy(string dclass, string dfunc, string sclass[, string sfunc])
	Copy a method from sclass to dclass as dfunc.  Source method is sfunc if given, dfunc otherwise. */
PHP_FUNCTION(classkit_method_copy)
{
	char *dclass, *dfunc, *sclass, *sfunc = NULL;
	int dclass_len, dfunc_len, sclass_len, sfunc_len = 0;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sss|s",	&dclass,	&dclass_len,
																	&dfunc,		&dfunc_len,
																	&sclass,	&sclass_len,
																	&sfunc,		&sfunc_len) == FAILURE) {
		RETURN_FALSE;
	}

	if (!sfunc) {
		sfunc = dfunc;
		sfunc_len = dfunc_len;
	}

	RETURN_BOOL(php_classkit_method_copy(dclass, dclass_len, dfunc, dfunc_len,
										 sclass, sclass_len, sfunc, sfunc_len TSRMLS_CC) == SUCCESS ? 1 : 0);
}
/* }}} */

/* {{{ proto array classkit_aggregate_methods(string dclass, string sclass[, long flags[, mixed prefix_or_list]])
	Copy all (or some) methods from sclass to dclass (overloaded)
 */
PHP_FUNCTION(classkit_aggregate_methods)
{
	char *dclass, *sclass;
	int dclass_len, sclass_len;
	long flags = 0;
	int override_target;
	zval *param = NULL;
	zend_class_entry *dce, *sce, *scope;
	int argc = ZEND_NUM_ARGS();
	char *key, *altkey = NULL;
	int key_len, altkey_len, altkey_maxlen = 0;
	long idx;

	if (zend_parse_parameters(argc TSRMLS_CC, "ss|bz",  &dclass, &dclass_len, &sclass, &sclass_len, &flags, &param) == FAILURE) {
		RETURN_FALSE;
	}

	if (php_classkit_fetch_class(dclass, dclass_len, &dce TSRMLS_CC) == FAILURE) {
		RETURN_FALSE;
	}

	if (php_classkit_fetch_class(sclass, sclass_len, &sce TSRMLS_CC) == FAILURE) {
		RETURN_FALSE;
	}

	override_target = flags & CLASSKIT_AGGREGATE_OVERRIDE;

	if (!param || Z_TYPE_P(param) == IS_STRING) {
		/* Aggregate everything in the source class, optionally prefixing it */
		if (param) {
			altkey_maxlen = Z_STRLEN_P(param) + 64;
			altkey = emalloc(altkey_maxlen);
			memcpy(altkey, Z_STRVAL_P(param), Z_STRLEN_P(param));
			altkey_len = Z_STRLEN_P(param);
		}
		
		array_init(return_value);
		zend_hash_internal_pointer_reset(&sce->function_table);
		while (zend_hash_get_current_key_ex(&sce->function_table, &key, &key_len, &idx, 0, NULL) == HASH_KEY_IS_STRING) {
			zend_function *dfe = NULL;
			if (altkey) {
				/* Needing to grow the buffer should be a rare event */
				if (altkey_len + key_len + 1 > altkey_maxlen) {
					altkey_maxlen = altkey_len + key_len + 33;
					altkey = erealloc(altkey, altkey_maxlen);
				}
				memcpy(altkey + Z_STRLEN_P(param), key, key_len);
				key_len += Z_STRLEN_P(param);
				altkey[key_len] = '\0';
				key = altkey;
			}

			if (override_target && zend_hash_find(&dce->function_table, key, key_len, (void**)&dfe) == SUCCESS) {
				if (php_classkit_check_call_stack(&dfe->op_array TSRMLS_CC) == FAILURE) {
					php_error_docref(NULL TSRMLS_CC, E_WARNING, "Cannot override active method %s::%s(). Skipping.", dclass, key);
					zend_hash_move_forward(&sce->function_table);
					continue;
				}

				scope = php_classkit_locate_scope(dce, dfe, key, key_len - 1);
				zend_hash_apply_with_arguments(EG(class_table), (apply_func_args_t)php_classkit_clean_children, 4, scope, dce, key, key_len - 1);
				if (zend_hash_del(&dce->function_table, key, key_len) == FAILURE) {
					php_error_docref(NULL TSRMLS_CC, E_WARNING, "Error removing old method in destination class %s::%s", dclass, key);
					zend_hash_move_forward(&sce->function_table);
					continue;
				}
			}

			if (php_classkit_method_copy(dclass, dclass_len, key, key_len - 1, sclass, sclass_len, 
										 key     + (param ? Z_STRLEN_P(param) : 0), 
										 key_len - (param ? Z_STRLEN_P(param) : 0) - 1 TSRMLS_CC) == SUCCESS) {
				add_next_index_stringl(return_value, key + (param ? Z_STRLEN_P(param) : 0), key_len - (param ? Z_STRLEN_P(param) : 0) - 1, 1);
			}

			zend_hash_move_forward(&sce->function_table);
		}

		if (altkey) {
			efree(altkey);
		}
	} else if (Z_TYPE_P(param) == IS_ARRAY) {
		/* Aggregate everything in the source class that's listed in the parameter */
		zval **zsname;

		array_init(return_value);
		zend_hash_internal_pointer_reset(HASH_OF(param));
		while (zend_hash_get_current_data(HASH_OF(param), (void**)&zsname) == SUCCESS) {
			zend_function *dfe = NULL;

			if (!zsname || !*zsname || Z_TYPE_PP(zsname) != IS_STRING || Z_STRLEN_PP(zsname) == 0) {
				php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid method name in list");
				zend_hash_move_forward(HASH_OF(param));
				continue;
			}

			if (!zend_hash_exists(&sce->function_table, Z_STRVAL_PP(zsname), Z_STRLEN_PP(zsname) + 1)) {
				php_error_docref(NULL TSRMLS_CC, E_NOTICE, "Source method %s::%s() does not exist", sclass, Z_STRVAL_PP(zsname));
				zend_hash_move_forward(HASH_OF(param));
				continue;
			}

			if (zend_hash_get_current_key_ex(HASH_OF(param), &key, &key_len, &idx, 0, NULL) != HASH_KEY_IS_STRING) {
				key = Z_STRVAL_PP(zsname);
				key_len = Z_STRLEN_PP(zsname) + 1;
			}

			if (override_target && zend_hash_find(&dce->function_table, key, key_len, (void**)&dfe) == SUCCESS) {
				if (php_classkit_check_call_stack(&dfe->op_array TSRMLS_CC) == FAILURE) {
					php_error_docref(NULL TSRMLS_CC, E_WARNING, "Cannot override active method %s::%s(). Skipping.", dclass, key);
					zend_hash_move_forward(HASH_OF(param));
					continue;
				}

				scope = php_classkit_locate_scope(dce, dfe, key, key_len - 1);
				zend_hash_apply_with_arguments(EG(class_table), (apply_func_args_t)php_classkit_clean_children, 4, scope, dce, key, key_len - 1);
				if (zend_hash_del(&dce->function_table, key, key_len) == FAILURE) {
					php_error_docref(NULL TSRMLS_CC, E_WARNING, "Error removing old method in destination class %s::%s", dclass, key);
					zend_hash_move_forward(HASH_OF(param));
					continue;
				}
			}

			if (php_classkit_method_copy(dclass, dclass_len, key, key_len - 1,
										 sclass, sclass_len, Z_STRVAL_PP(zsname), Z_STRLEN_PP(zsname) TSRMLS_CC) == SUCCESS) {
				add_next_index_stringl(return_value, Z_STRVAL_PP(zsname), Z_STRLEN_PP(zsname), 1);
			}

			zend_hash_move_forward(HASH_OF(param));
		}
	} else {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Parameter four expects STRING or ARRAY");
		RETURN_FALSE;
	}
}
/* }}} */

/* {{{ array classkit_import(string filename)
	Import new class method definitions from a file */
PHP_FUNCTION(classkit_import)
{
	zval *filename;
	zend_op_array *new_op_array;
	int original_num_classes = zend_hash_num_elements(EG(class_table));
	int original_num_functions = zend_hash_num_elements(EG(function_table));
	char *key;
	int key_len, type, i;
	long idx;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "z", &filename) == FAILURE) {
		RETURN_FALSE;
	}
	convert_to_string(filename);

	new_op_array = compile_filename(ZEND_INCLUDE, filename TSRMLS_CC);
	if (!new_op_array) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Import Failure");
		RETURN_FALSE;
	}
	/* We never really needed the main loop opcodes to begin with */
	destroy_op_array(new_op_array TSRMLS_CC);
	efree(new_op_array);

	/* Get rid of functions, we're not importing them (though maybe someday....) */
	zend_hash_internal_pointer_end(EG(function_table));
	for(i = zend_hash_num_elements(EG(function_table));
		i > original_num_functions; i--) {
		zend_function *fe = NULL;

		zend_hash_get_current_data(EG(function_table), (void**)&fe);
		if (((type = zend_hash_get_current_key_ex(EG(function_table), &key, &key_len, &idx, 0, NULL)) != HASH_KEY_NON_EXISTANT) && 
			fe && fe->type == ZEND_USER_FUNCTION) {
			if (type == HASH_KEY_IS_STRING) {
				if (zend_hash_del(EG(function_table), key, key_len) == FAILURE) {
					php_error_docref(NULL TSRMLS_CC, E_ERROR, "Inconsistency cleanup up import environment");
					RETURN_FALSE;
				}
			} else {
				if (zend_hash_index_del(EG(function_table), idx) == FAILURE) {
					php_error_docref(NULL TSRMLS_CC, E_ERROR, "Inconsistency cleanup up import environment");
					RETURN_FALSE;
				}
			}
		} else {
			php_error_docref(NULL TSRMLS_CC, E_ERROR, "Inconsistency cleaning up import environment");
			RETURN_FALSE;
		}
		zend_hash_move_backwards(EG(function_table));
	}

	array_init(return_value);

	/* Pop the "new" classes off the class table */
	zend_hash_internal_pointer_end(EG(class_table));
	for(i = zend_hash_num_elements(EG(class_table));
		i > original_num_classes; i--) {
		zend_class_entry *ce = NULL;

		zend_hash_get_current_data(EG(class_table), (void**)&ce);
#ifdef ZEND_ENGINE_2
		if (ce) {
			ce = *((zend_class_entry**)ce);
		}
#endif

		if (((type = zend_hash_get_current_key_ex(EG(class_table), &key, &key_len, &idx, 0, NULL)) != HASH_KEY_NON_EXISTANT) && 
			ce && ce->type == ZEND_USER_CLASS) {
			zend_class_entry *dce;
			zend_function *fe;
			char *importbuf;
			int importbuf_len;

			if (ce->name_length == key_len - 1) {
				/* If their lengths are the same it's all the evidence we need that their names are the same.
					If their names are the same its because this class was freshly declared, no need for aggregation. */
				importbuf_len = spprintf(&importbuf, 0, "%s::*", ce->name);
				add_next_index_stringl(return_value, importbuf, importbuf_len, 0);
				continue;
			}
			/* Otherwise aggregate methods from temp class to real one then demolish it */

			if (php_classkit_fetch_class(ce->name, ce->name_length, &dce TSRMLS_CC) == FAILURE) {
				/* Oddly non-existant target class or error retreiving it... Or it's an internal class... */
				continue;
			}
			
			zend_hash_internal_pointer_reset(&ce->function_table);
			while (zend_hash_get_current_data(&ce->function_table, (void**)&fe) == SUCCESS) {
				zend_function *dfe;
				int fn_len;
				zend_class_entry *fe_scope;
				
				fn_len = strlen(fe->common.function_name);
				fe_scope = php_classkit_locate_scope(ce, fe, fe->common.function_name, fn_len);
				
				if (fe_scope != ce) {
					/* This is an inhereted function, let's skip it */
					zend_hash_move_forward(&ce->function_table);
					continue;
				}

				if (zend_hash_find(&dce->function_table, fe->common.function_name, fn_len + 1, (void**)&dfe) == SUCCESS) {
					if (php_classkit_check_call_stack(&dfe->op_array TSRMLS_CC) == FAILURE) {
						php_error_docref(NULL TSRMLS_CC, E_WARNING, "Cannot override active method %s::%s(). Skipping.", dce->name, fe->common.function_name);
						zend_hash_move_forward(&ce->function_table);
						continue;
					}

					zend_class_entry *scope = php_classkit_locate_scope(dce, dfe, fe->common.function_name, fn_len);
					zend_hash_apply_with_arguments(EG(class_table), (apply_func_args_t)php_classkit_clean_children, 4, scope, dce, fe->common.function_name, fn_len);
					if (zend_hash_del(&dce->function_table, fe->common.function_name, fn_len + 1) == FAILURE) {
						php_error_docref(NULL TSRMLS_CC, E_WARNING, "Error removing old method in destination class %s::%s", dce->name, fe->common.function_name);
						zend_hash_move_forward(&ce->function_table);
						continue;
					}
				}

				function_add_ref(fe);
#ifdef ZEND_ENGINE_2
				fe->common.scope = dce;
#endif
				zend_hash_apply_with_arguments(EG(class_table), (apply_func_args_t)php_classkit_update_children, 5, dce, dce, fe, fe->common.function_name, fn_len);

				if (zend_hash_add(&dce->function_table, fe->common.function_name, fn_len + 1, fe, sizeof(zend_function), NULL) == FAILURE) {
					php_error_docref(NULL TSRMLS_CC, E_WARNING, "Failure importing %s::%s()", ce->name, fe->common.function_name);
					zend_hash_move_forward(&ce->function_table);
					continue;
				}

				importbuf_len = spprintf(&importbuf, 0, "%s::%s", ce->name, fe->common.function_name);
				add_next_index_stringl(return_value, importbuf, importbuf_len, 0);

				zend_hash_move_forward(&ce->function_table);
			}

			if (type == HASH_KEY_IS_STRING) {
				if (zend_hash_del(EG(class_table), key, key_len) == FAILURE) {
					php_error_docref(NULL TSRMLS_CC, E_ERROR, "Inconsistency cleaning up import environment");
					zval_dtor(return_value);
					RETURN_FALSE;
				}
			} else {
				if (zend_hash_index_del(EG(class_table), idx) == FAILURE) {
					php_error_docref(NULL TSRMLS_CC, E_ERROR, "Inconsistency cleaning up import environment");
					zval_dtor(return_value);
					RETURN_FALSE;
				}
			}
		} else {
			php_error_docref(NULL TSRMLS_CC, E_ERROR, "Inconsistency cleaning up import environment");
			zval_dtor(return_value);
			RETURN_FALSE;
		}
		zend_hash_move_backwards(EG(class_table));
	}

}
/* }}} */

/* {{{ PHP_MINIT_FUNCTION
 */
PHP_MINIT_FUNCTION(classkit)
{
#ifdef ZEND_ENGINE_2
	REGISTER_LONG_CONSTANT("CLASSKIT_ACC_PUBLIC",			ZEND_ACC_PUBLIC,				CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("CLASSKIT_ACC_PROTECTED",		ZEND_ACC_PROTECTED,				CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("CLASSKIT_ACC_PRIVATE",			ZEND_ACC_PRIVATE,				CONST_CS | CONST_PERSISTENT);
#endif
	REGISTER_STRING_CONSTANT("CLASSKIT_VERSION",			CLASSKIT_VERSION,				CONST_CS | CONST_PERSISTENT);

	REGISTER_LONG_CONSTANT("CLASSKIT_AGGREGATE_OVERRIDE",	CLASSKIT_AGGREGATE_OVERRIDE,	CONST_CS | CONST_PERSISTENT);

	return SUCCESS;
}
/* }}} */

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
